class ThingDropped(Exception):
    """Raise when pipeline dropped a thing"""


class FiledValidationError(Exception):
    pass


class ItemExtractError(Exception):
    """Raise when extract item when error"""



"""The thing`s usage is simple, can be created by ant, processed by ants and pipelines"""
from typing import Any, Optional, Iterator, Tuple, Dict, Type, Union, List, DefaultDict
from collections.abc import MutableMapping
import json
import abc
from collections import defaultdict
import logging
import re

from lxml import html
from yarl import URL
import jpath

from .exceptions import FiledValidationError, ItemExtractError


class Request:
    __slots__ = ('url', 'params', 'method', 'headers', 'cookies', 'data', 'proxy', 'max_redirects', 'allow_redirects')

    def __init__(self, url: Union[str, URL], method='GET', params: Optional[dict]=None, headers: Optional[dict]=None,
                 cookies: Optional[dict]=None, data: Optional[Any]=None, proxy: Optional[str]=None,
                 allow_redirects=True, max_redirects=10):
        if isinstance(url, str):
            url = URL(url)
        self.url = url
        self.params = params
        self.method = method
        self.headers = headers
        self.cookies = cookies
        self.data = data
        self.proxy = proxy
        self.allow_redirects = allow_redirects
        self.max_redirects = max_redirects

    def __repr__(self):
        return '{:s}: {:s} {:s}'.format(self.__class__.__name__, self.method, str(self.url))


class Response:
    __slots__ = ('url', 'status', 'headers', 'cookies', 'encoding', 'content', '_text', 'request', '_html_element',
                 '_json')

    def __init__(self, request: Request, status: int, content: bytes, headers: Optional[dict]=None,
                 cookies: Optional[dict]=None, encoding: str='utf-8'):
        self.request = request
        self.url = request.url
        self.status = status
        self.headers = headers
        self.cookies = cookies
        self.encoding = encoding
        self.content = content
        self._text = None
        self._html_element = None
        self._json = None

    @property
    def text(self) -> str:
        if self._text is None:
            self._text = self.content.decode(self.encoding)
        return self._text

    @property
    def json(self) -> Any:
        if self._json is None:
            self._json = json.loads(self.text)
        return self._json

    @property
    def html_element(self) -> html.HtmlElement:
        if self._html_element is None:
            self._html_element = html.fromstring(self.text)
        return self._html_element

    def __repr__(self):
        return '{:s}: {:s} {:d}'.format(self.__class__.__name__, str(self.url), self.status)


class IntField:
    _type = int
    storage_name = ''
    __shadow_name_prefix = '__field#'

    def __init__(self, null: bool=False):
        self.null = null

    def __set__(self, instance: 'Item', value: Any) -> None:
        setattr(instance, self.storage_name, value)

    def __get__(self, instance: 'Item', owner: Type['Item']) -> Any:
        return getattr(instance, self.storage_name)

    def __delete__(self, instance):
        delattr(instance, self.storage_name)

    def validate(self, value: Any) -> Any:
        """:raise FiledValidationError"""
        if self.null and value is None:
            return value
        elif value is not None:
            try:
                return self._type(value)
            except (ValueError, TypeError) as e:
                raise FiledValidationError(e)
        else:
            raise FiledValidationError('value is None')

    @classmethod
    def make_shadow_name(cls, name: str) -> str:
        if name == cls.__shadow_name_prefix:
            raise AttributeError('This name: {:s} has been used internally'.format(cls.__shadow_name_prefix))
        return cls.__shadow_name_prefix + name

    @classmethod
    def is_shadow_name(cls, name: str) -> bool:
        return cls.__shadow_name_prefix in name

    @classmethod
    def get_name_from_shadow(cls, name: str) -> str:
        return name.replace(cls.__shadow_name_prefix, '')


class FloatField(IntField):
    _type = float


class StringField(IntField):
    _type = str


class BytesField(IntField):
    _type = bytes


class ItemMeta(abc.ABCMeta):
    def __init__(cls, name: str, bases: Tuple[type, ...], attr_dict: Dict[str, Any]):
        super().__init__(name, bases, attr_dict)
        for k, v in attr_dict.items():
            if isinstance(v, IntField):
                v.storage_name = IntField.make_shadow_name(k)


class Item(MutableMapping, metaclass=ItemMeta):
    _is_validating = False

    def __setattr__(self, key: str, value: Any) -> None:
        if isinstance(key, str):
            super().__setattr__(key, value)
        else:
            raise AttributeError('The key`s type must be str')

    def __setitem__(self, key: str, value: Any) -> None:
        try:
            setattr(self, key, value)
        except AttributeError:
            raise KeyError(key)

    def __getitem__(self, key: str) -> Any:
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(key)

    def __delitem__(self, key: str) -> None:
        try:
            delattr(self, key)
        except AttributeError:
            raise KeyError(key)

    def __len__(self) -> int:
        return len(self.__dict__)

    def __iter__(self) -> Iterator[str]:
        keys = []
        for k in self.__dict__:
            if IntField.is_shadow_name(k):
                keys.append(IntField.get_name_from_shadow(k))
            else:
                keys.append(k)
        return iter(keys)

    def validate(self):
        """Validate item`s type.
        Get descriptors reference method from __class__.__dict__"""
        class_dict = self.__class__.__dict__
        for k, obj in class_dict.items():
            if isinstance(obj, IntField):
                try:
                    setattr(self, k, class_dict[k].validate(getattr(self, k, None)))
                except FiledValidationError as e:
                    if 'value is None' in str(e):
                        raise FiledValidationError('{:s}`s value is None'.format(k))
                    else:
                        raise e

    def __repr__(self):
        return '{:s}: {:s}'.format(self.__class__.__name__, str(dict(self)))


Things = Union[Request, Response, Item]


class ItemExtractor:
    take_first = 'take_first'
    join_all = 'join_all'
    do_nothing = 'do_nothing'

    def __init__(self, item_class: Type[Item]):
        self.item_class = item_class
        self.logger = logging.getLogger(self.__class__.__name__)
        self.path = defaultdict(list)  # type: DefaultDict[str, List[Tuple[str, str, str]]]

    def add_xpath(self, key: str, xpath: str, extract_type=take_first):
        self.path[key].append(('xpath', xpath, extract_type))

    def add_regex(self, key: str, pattern: str, extract_type=take_first):
        self.path[key].append(('regex', pattern, extract_type))

    def add_jpath(self, key: str, jpath, extract_type=take_first):
        self.path[key].append(('jpath', jpath, extract_type))

    def extract(self, response: Response) -> Item:
        """Extract item from response by path with xpath, jpath or re."""
        self.logger.debug('Extract item: {:s} with path: {:s}'.format(self.item_class.__name__, str(self.path)))
        item = self.item_class()
        for key, all_xpath in self.path.items():
            value = None
            for path_type, path, extract_type in all_xpath:
                if path_type == 'xpath':
                    extract_value = response.html_element.xpath(path)
                elif path_type == 'regex':
                    extract_value = re.findall(path, response.text)
                elif path_type == 'jpath':
                    extract_value = jpath.get_all(path, response.json)
                else:
                    continue
                if len(extract_value) == 0:
                    continue
                # handle by extract type
                if extract_type == self.take_first:
                    extract_value = extract_value[0]
                elif extract_type == self.join_all:
                    extract_value = list(filter(lambda x: isinstance(x, str), extract_value))  # join string only
                    extract_value = ''.join(extract_value)
                elif extract_value == self.do_nothing:
                    pass
                # check multiple path`s result
                if extract_value is not None:
                    if value is not None and value != extract_value:
                        raise ItemExtractError(
                            'Match different result: {:s} and {:s} for key: {:s}'.format(value, extract_value, key))
                    elif extract_value is not None:
                        value = extract_value
            item[key] = value
        return item



from typing import Any, Optional, Iterator, Generator, List, Coroutine, Union, Dict
import abc
import itertools
import logging
import asyncio
from asyncio.queues import Queue, QueueEmpty
from itertools import islice
import async_timeout

import aiohttp
from aiohttp.client_reqrep import ClientResponse
from aiohttp import ClientSession
from yarl import URL

from .pipelines import Pipeline
from .things import Request, Response, Item, Things
from .exceptions import ThingDropped


DEFAULT_VALUE = -1


class Ant(abc.ABC):
    response_pipelines = [Pipeline()]  # type: List[Pipeline]
    request_pipelines = [Pipeline()]  # type: List[Pipeline]
    item_pipelines = [Pipeline()]  # type: List[Pipeline]

    CONCURRENT_LIMIT = 30  # backgroud coroutines concurrent limit
    COROUTINE_TIMEOUT = 180

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.sessions = {}  # type: Dict[str, ClientSession]
        # for background coroutines - coroutines schedule by self.ensure_future method
        self.__queue = Queue()  # background coroutines waiting for execution
        self.__done_queue = Queue()  # execution completed background coroutines
        self.__count = 0  # backgroud coroutines be executed
        self.__done_count = 0  # backgroud coroutines completed

    async def request(self, url: Union[str, URL], method='GET', params: Optional[dict]=None,
                      headers: Optional[dict]=None, cookies: Optional[dict]=None, data: Optional[Any]=None,
                      proxy: Optional[str]=None, timeout: int=DEFAULT_VALUE) -> Response:
        self.logger.debug('{:s} {:s}'.format(method, str(url)))
        kwargs = locals()
        kwargs.pop('self')
        timeout = kwargs.pop('timeout')

        req = Request(**kwargs)
        req = await self.timeout_wrapper(self._handle_thing_with_pipelines(req, self.request_pipelines), timeout=timeout)

        res = await self.timeout_wrapper(self._request(req), timeout=timeout)

        res = await self.timeout_wrapper(self._handle_thing_with_pipelines(res, self.response_pipelines),
                                         timeout=timeout)
        return res

    async def collect(self, item: Item, timeout: int=DEFAULT_VALUE) -> None:
        self.logger.debug('Collect item: ' + str(item))
        await self.timeout_wrapper(self._handle_thing_with_pipelines(item, self.item_pipelines), timeout=timeout)

    async def open(self, timeout: int=DEFAULT_VALUE) -> None:
        self.logger.info('Opening')
        for pipeline in itertools.chain(self.item_pipelines, self.response_pipelines, self.request_pipelines):
            obj = pipeline.on_spider_open(self)
            if isinstance(obj, Coroutine):
                await self.timeout_wrapper(obj, timeout=timeout)

    async def close(self, timeout: int=DEFAULT_VALUE) -> None:
        for pipeline in itertools.chain(self.item_pipelines, self.response_pipelines, self.request_pipelines):
            obj = pipeline.on_spider_close(self)
            if isinstance(obj, Coroutine):
                await self.timeout_wrapper(obj, timeout=timeout)
        for session in self.sessions.values():
            await session.close()
        self.logger.info('Closed')

    @abc.abstractmethod
    async def run(self) -> None:
        """App custom entrance"""

    async def main(self) -> None:
        try:
            await self.open()
            await self.run()
            await self.__run_until_complete()
            await self.close()
        except Exception as e:
            self.logger.exception('Run main coroutine with ' + e.__class__.__name__)

    def ensure_future(self, coroutine: Coroutine, timeout: int=DEFAULT_VALUE) -> None:
        """Custom ensure_future method provide coroutines concurrent limit and watch dog"""
        if timeout == DEFAULT_VALUE:
            timeout = self.COROUTINE_TIMEOUT

        def _done_callback(f):
            exception = f.exception()
            if exception is not None:
                try:
                    raise exception
                except exception.__class__:
                    self.logger.exception(exception)

            self.__done_count += 1
            self.__done_queue.put_nowait(f)
            try:
                running_count = self.__count - self.__done_count
                if running_count < self.CONCURRENT_LIMIT:
                    next_coroutine = self.__queue.get_nowait()
                    self.__count += 1
                    asyncio.ensure_future(next_coroutine).add_done_callback(_done_callback)
            except QueueEmpty:
                pass

        running_count = self.__count - self.__done_count
        if running_count < self.CONCURRENT_LIMIT:
            self.__count += 1
            asyncio.ensure_future(self.timeout_wrapper(coroutine, timeout=timeout)).add_done_callback(_done_callback)
        else:
            self.__queue.put_nowait(self.timeout_wrapper(coroutine, timeout=timeout))

    def as_completed(self, coroutines: Union[Iterator[Coroutine], List[Coroutine]],
                     limit: int=DEFAULT_VALUE, timeout: int=DEFAULT_VALUE) -> Generator[Coroutine, None, None]:
        """Custom as_completed method provide coroutines concurrent limit"""
        if limit == DEFAULT_VALUE:
            limit = self.CONCURRENT_LIMIT
        if timeout == DEFAULT_VALUE:
            timeout = self.COROUTINE_TIMEOUT

        if isinstance(coroutines, List):
            coroutines = iter(coroutines)

        queue = Queue()
        todo = []

        def _done_callback(f):
            queue.put_nowait(f)
            todo.remove(f)
            try:
                nf = asyncio.ensure_future(next(coroutines))
                nf.add_done_callback(_done_callback)
                todo.append(nf)
            except StopIteration:
                pass

        async def _wait_for_one():
            f = await queue.get()
            return f.result()

        if limit <= 0:
            fs = {asyncio.ensure_future(self.timeout_wrapper(cor, timeout=timeout)) for cor in coroutines}
        else:
            fs = {asyncio.ensure_future(self.timeout_wrapper(cor, timeout=timeout))
                  for cor in islice(coroutines, 0, limit)}
        for f in fs:
            f.add_done_callback(_done_callback)
            todo.append(f)

        while len(todo) > 0:
            yield _wait_for_one()

    async def timeout_wrapper(self, coroutine: Coroutine, timeout: int=DEFAULT_VALUE):
        if timeout == DEFAULT_VALUE:
            timeout = self.COROUTINE_TIMEOUT
        if timeout > 0:
            with async_timeout.timeout(timeout):
                return await coroutine
        else:
            return await coroutine

    async def __run_until_complete(self) -> None:
        """Watch dog,  wait all background coroutines to be completed"""
        while self.__done_count != self.__count:
            await self.__done_queue.get()

    async def _handle_thing_with_pipelines(self, thing: Things, pipelines: List[Pipeline]) -> Optional[Things]:
        """Process thing one by one, break the process chain when get None
        :raise ThingProcessError"""
        self.logger.debug('Process thing: ' + str(thing))
        raw_thing = thing
        for pipeline in pipelines:
            thing = pipeline.process(self, thing)
            if isinstance(thing, Coroutine):
                thing = await thing
            if thing is None:
                msg = 'The thing {:s} is dropped by {:s}'.format(str(raw_thing),
                                                                 pipeline.__class__.__name__)
                self.logger.warning(msg)
                raise ThingDropped(msg)
        return thing

    async def _request(self, req: Request) -> Response:
        kwargs = {k: getattr(req, k) for k in req.__slots__}
        cookies = kwargs.pop('cookies')

        # proxy auth not work in one session with many requests
        if req.proxy is not None and URL(req.proxy).user is not None:
            async with aiohttp.ClientSession(cookies=cookies) as session:
                async with session.request(**kwargs) as aio_response:
                    await aio_response.read()
                return self._convert_response(aio_response, req)

        host = req.url.host
        if host not in self.sessions:
            session = aiohttp.ClientSession(cookies=cookies)
            self.sessions[host] = session
        else:
            session = self.sessions[host]
            if cookies is not None:
                session.cookie_jar.update_cookies(cookies)

        async with session.request(**kwargs) as aio_response:
            await aio_response.read()
        return self._convert_response(aio_response, req)

    @staticmethod
    def _convert_response(aio_response: ClientResponse, request: Request) -> Response:
        return Response(request, aio_response.status, aio_response._content,
                        headers=aio_response.headers, cookies=aio_response.cookies,
                        encoding=aio_response._get_encoding())



from typing import Optional, List, Tuple, DefaultDict, Dict, Any
import logging
from collections import defaultdict
import json
import asyncio
import os
import time
import datetime

import aiomysql

from .things import Things, Response, Request, Item
from .exceptions import FiledValidationError


class Pipeline:
    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    async def on_spider_open(self, ant: '.ant_nest.Ant') -> None:
        """Call when ant open, method or coroutine"""

    async def on_spider_close(self, ant: '.ant_nest.Ant') -> None:
        """Call when ant close, method or coroutine"""

    async def process(self, ant: '.ant_nest.ant', thing: Things) -> Optional[Things]:
        """method or coroutine"""
        return thing


class ReportPipeline(Pipeline):
    def __init__(self):
        self.count = 0
        self.report_type = None
        self.last_time = time.time()
        self.last_count = 0
        super().__init__()

    def process(self, ant: '.ant_nest.ant', thing: Things) -> Things:
        if self.report_type is None:
            self.report_type = thing.__class__.__name__
        self.count += 1
        now_time = time.time()
        if now_time - self.last_time > 60:
            count = self.count - self.last_count
            self.logger.info('Get {:d} {:s} in total with {:d}/min rate'.format(self.count, self.report_type, count))
            self.last_time = now_time
            self.last_count = self.count
        return thing

    def on_spider_close(self, ant: '.ant_nest.Ant'):
        if self.report_type is not None:
            self.logger.info('Get {:d} {:s} in total'.format(self.count, self.report_type))


# Response pipelines
class ResponseFilterErrorPipeline(Pipeline):
    def process(self, ant: '.ant_nest.Ant', thing: Response) -> Optional[Response]:
        if thing.status >= 400:
            self.logger.warning('Response: {:s} has bean dropped'.format(str(thing)))
            return None
        else:
            return thing


class ResponseRetryPipeline(Pipeline):
    """This pipeline should be in front of the pipeline chain"""
    def __init__(self, retries: int=3, interval: int=3):
        self.retries = retries
        self.interval = interval
        super().__init__()

    async def process(self, ant: '.ant_nest.ant', thing: Response) -> Optional[Response]:
        retries = self.retries
        while retries >= 0:
            if thing.status >= 400:
                self.logger.info('Retry {:s} because http status {:d}'.format(str(thing.request), thing.status))
                await asyncio.sleep(self.interval)
                thing = await ant._request(thing.request)
            else:
                return thing
            retries -= 1
        self.logger.warning('{:d} reties failed for {:s}'.format(self.retries, str(thing)))
        return thing


# Request pipelines
class RequestNoRedirectsPipeline(Pipeline):
    def process(self, ant: '.ant_nest.Ant', thing: Request) -> Request:
        thing.allow_redirects = False
        thing.max_redirects = 0
        return thing


class RequestProxyPipeline(Pipeline):
    def __init__(self, proxy: str):
        self.proxy = proxy
        super().__init__()

    def process(self, ant: '.ant_nest.ant', thing: Request) -> Request:
        thing.proxy = self.proxy
        return thing


class RequestDuplicateFilterPipeline(Pipeline):
    def __init__(self):
        self.__request_urls = set()
        super().__init__()

    def process(self, ant: '.ant_nest.ant', thing: Request) -> Optional[Request]:
        if thing.url in self.__request_urls:
            return None
        else:
            self.__request_urls.add(thing.url)
            return thing


class RequestUserAgentPipeline(Pipeline):
    user_agent = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_1) AppleWebKit/537.36 (KHTML, like Gecko)' \
                 'Chrome/62.0.3202.89 Safari/537.36 Name'

    def __init__(self, user_agent=user_agent):
        super().__init__()
        self.user_agent = user_agent

    def process(self, ant, thing):
        if thing.headers is None:
            thing.headers = {}
        thing.headers['User-Agent'] = self.user_agent
        return thing


# Item pipelines
class ItemPrintPipeline(Pipeline):
    def process(self, ant: '.ant_nest.ant', thing: Item) -> Item:
        self.logger.info(str(thing))
        return thing


class ItemValidatePipeline(Pipeline):
    def process(self, ant: '.ant_nest.ant', thing: Item) -> Optional[Item]:
        try:
            thing.validate()
            return thing
        except FiledValidationError:
            return None


class ItemFieldReplacePipeline(Pipeline):
    def __init__(self, fields: List[str], excess_chars: Tuple[str]=('\r', '\n', '\t')):
        self.fields = fields
        self.excess_chars = excess_chars
        super().__init__()

    def process(self, ant: '.ant_nest.ant', thing: Item) -> Item:
        for field in self.fields:
            for char in self.excess_chars:
                if field in thing and isinstance(thing[field], str):
                    thing[field] = thing[field].replace(char, '')
        return thing


class ItemJsonDumpPipeline(Pipeline):
    def __init__(self, file_dir: str='.'):
        super().__init__()
        self.file_dir = file_dir
        self.data = defaultdict(list)  # type: DefaultDict[str, List[Dict]]

    def process(self, ant: '.ant_nest.ant', thing: Item) -> Item:
        self.data[thing.__class__.__name__].append(dict(thing))
        return thing

    def dump(self, file_path: str, data: dict):
        with open(file_path, 'w') as f:
            json.dump(data, f)

    async def on_spider_close(self, ant: '.ant_nest.Ant'):
        for file_name, data in self.data.items():
            ant.ensure_future(
                asyncio.get_event_loop().run_in_executor(
                    None, self.dump, os.path.join(self.file_dir, file_name + '.json'), data))


class ItemBaseMysqlPipeline(Pipeline):
    def __init__(self, host: str, port: int, user: str, password: str, database: str, table: str, charset: str='utf8'):
        super().__init__()
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database
        self.table = table
        self.charset = charset
        self.pool = None

    async def on_spider_open(self, ant: '.ant_nest.ant') -> None:
        self.pool = await aiomysql.create_pool(host=self.host, port=self.port, user=self.user, password=self.password,
                                               db=self.database, charset=self.charset, use_unicode=True)

    def on_spider_close(self, ant: '.ant_nest.ant') -> None:
        self.pool.close()

    async def push_data(self, sql: str) -> None:
        self.logger.debug('Executing SQL: ' + sql)
        async with self.pool.get() as conn:
            try:
                async with conn.cursor() as cur:
                    await cur.execute(sql)
                await conn.commit()
            except Exception as e:
                self.logger.exception('Push data with ' + e.__class__.__name__)

    @staticmethod
    def convert_item_value(value: Any) -> str:
        """Parse value to str for making sql string, eg: False -> '0'"""
        if isinstance(value, bool):
            return '1' if value else '0'
        elif isinstance(value, str):
            value = value.replace('\"', '\\\"')
            return '"{:s}"'.format(value)
        elif isinstance(value, datetime.datetime):
            return '"{:s}"'.format(str(value))
        elif isinstance(value, datetime.date):
            return '"{:s}"'.format(str(value))
        elif value is None:
            return 'null'
        else:
            return str(value)


class ItemMysqlInsertPipeline(ItemBaseMysqlPipeline):
    sql_format = 'INSERT IGNORE INTO {database}.{table} ({fields}) VALUES ({values})'

    async def process(self, ant: '.ant_nest.ant', thing: Item):
        fields = []
        values = []
        for k, v in thing.items():
            fields.append(k)
            values.append(self.convert_item_value(v))
        sql = self.sql_format.format(database=self.database, table=self.table, fields=','.join(fields),
                                     values=','.join(values))
        await self.push_data(sql)
        return thing


class ItemMysqlUpdatePipeline(ItemBaseMysqlPipeline):
    sql_format = 'UPDATE {database}.{table} SET {pairs} WHERE {primary_key}={primary_value}'

    def __init__(self, primary_key: str, host: str, port: int, user: str, password: str, database: str, table: str,
                 charset: str='utf8'):
        super().__init__(host, port, user, password, database, table, charset=charset)
        self.primary_key = primary_key

    async def process(self, ant: '.ant_nest.ant', thing: Item) -> Item:
        pairs = []
        primary_value = None

        for k, v in thing.items():
            if k == self.primary_key:
                primary_value = self.convert_item_value(v)
            else:
                pairs.append('{:s}={:s}'.format(k, self.convert_item_value(v)))

        if primary_value is not None:
            sql = self.sql_format.format(database=self.database, table=self.table, pairs=','.join(pairs),
                                         primary_key=self.primary_key, primary_value=primary_value)
            await self.push_data(sql)
        return thing
